/************************************************************************
 *
 * Licensed Materials - Property of IBM
 * Program Name: IBM DB2Everyplace Sync Client
 * (c ) Copyright IBM Corp. 1999-2003. All Rights Reserved.
 *
 * US Government Users Restricted Rights 
 * Use, duplication or disclosure restricted by GSA ADP Schedule
 * Contract with IBM Corp. Licensed Materials - Property of IBM
 *
 * IBM grants you ("Licensee") a non-exclusive, royalty free, license to
 * use Sample Code.  This software is provided as-is and without any
 * warranties, whether EXPRESS OR  IMPLIED, INCLUDING ANY IMPLIED
 * WARRANTY OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE OR 
 * NON-INFRINGEMENT. IBM AND ITS LICENSORS SHALL NOT BE LIABLE FOR ANY 
 * DAMAGES SUFFERED BY LICENSEE THAT RESULT FROM YOUR USE OF THE SOFTWARE.
 * IN NO EVENT WILL IBM OR ITS LICENSORS BE LIABLE FOR ANY LOST REVENUE,
 * PROFIT OR DATA, OR FOR DIRECT, INDIRECT, SPECIAL, CONSEQUENTIAL, 
 * INCIDENTAL OR PUNITIVE DAMAGES, HOWEVER CAUSED AND REGARDLESS OF THE 
 * THEORY OF LIABILITY, ARISING OUT OF THE USE OF OR INABILITY TO USE 
 * SOFTWARE, EVEN IF IBM HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH 
 * DAMAGES.
 *
 ************************************************************************
 */

import java.sql.*;
import java.util.Properties;
import java.util.Hashtable;
import java.util.Enumeration;
import java.io.*;

import com.ibm.mobileservices.isync.ConflictReader;
import com.ibm.mobileservices.isync.ISync;
import com.ibm.mobileservices.isync.ISyncException;
import com.ibm.mobileservices.isync.event.ISyncEvent;
import com.ibm.mobileservices.isync.event.ISyncListener;

/**
	Implementor of the eventIssued methodwhich handles info, error,
	conflict and query types of events.
*/

public class DB2SyncConsoleListener implements ISyncListener 
{
	private Properties isyncProps;
	protected boolean retryMaxReached;
	protected PrintStream logWriter;

	/**
	 	Create an instance setting in the isyncProps used
		to set up the sync engine and supply info in the eventIssued
		method.

		@param isyncProps - sync engine/event properties
	*/

	public DB2SyncConsoleListener(Properties isyncProps) {
		super();
		this.isyncProps = isyncProps;
	}

	/**
		Handle events from the isync driver which can be informational,
		error, data conflict, or query for information types of events.

		@param evt the event containing type, code, object for various use, etc.
	*/

	public int eventIssued(ISyncEvent evt)
	{
		// In case the event is null;
		if (evt == null) 
		{
			logWrite("Severe Error event is null on the callbacks from the iSyncEngine.");
			return ISync.RTNCB_DEFAULT;
		}

		int evtType = evt.getEventType();
		int evtCode = evt.getEventCode();

		logISync(evt);

		switch (evtType) {
			// display event status
			case ISync.EVTTYPE_INFO :
			case ISync.EVTTYPE_ERROR :
				if (evtCode == ISync.EVT_INF_STATS)
				{
					String trace = isyncProps.getProperty("isync.trace");
					printStats(evt);
				}
				return ISync.RTNCB_DONE;

			case ISync.EVTTYPE_RETRY :
				if (evt.getNumberOfRetries() >= 10)
					retryMaxReached = true;

				if (retryMaxReached)			
					return ISync.RTNCB_REPLY_NO;
				else
					return ISync.RTNCB_REPLY_YES;

			case ISync.EVTTYPE_CONFLICT :
				ConflictReader cr = (ConflictReader)evt.getEventInfo();
				processRejectedData(cr);
				return ISync.RTNCB_DONE;

			case ISync.EVTTYPE_QUERY :
				if (evtCode == ISync.EVT_QUE_LOGIN)
				{
					String login[] = (String[]) evt.getEventInfo();
					logWrite("Querying login for local encryption for database: " + login[0]);

					if (isyncProps.getProperty("target.db.user") == null 
						|| isyncProps.getProperty("target.db.password") == null)
					{
						try
						{
							BufferedReader in =
								new BufferedReader(new InputStreamReader(System.in));
	
							String buf;
							System.out.print("Username: ");
							buf = in.readLine();
							login[1] =buf.trim();
							System.out.print("Password: ");
							buf = in.readLine();
							login[2] =buf.trim();
						}
						catch (IOException ioe)
						{
							System.out.println(ioe);
							logWrite(ioe.toString());
							logExceptionStack(ioe);
						}
					}
					else
					{
						login[1] = isyncProps.getProperty("target.db.user");
						login[2] = isyncProps.getProperty("target.db.password");
					}

					return ISync.RTNCB_DONE;
				}
				else if (evtCode == ISync.EVT_QUE_SUBSCRIPTION_TARGET)
				{
					// Specify multiple directories for sync (optional)
					int subsType = evt.getSubscriptionType();
					String subsName = evt.getSubscriptionName();
					String login[] = (String [])evt.getEventInfo();

					if (subsType == ISync.SUBSTYPE_FILE)
						login[0] = isyncProps.getProperty("filesubdir");
					else if (subsName != null &&
						subsName.toUpperCase().startsWith("DSYCLIENTSTAT"))
					{
						return ISync.RTNCB_DEFAULT;
					}
					else
						login[0] = isyncProps.getProperty("jdbcsuburl");

					System.out.println("Sample set subs " + login[0]);
					return ISync.RTNCB_DONE;
				}
				return ISync.RTNCB_DEFAULT;

				// ignore other event types
			default :
				break;
		}

		return ISync.RTNCB_DEFAULT;

	}

	/**
		Print sync statistics such as time, number of object synchronized,
		names of tables or files, etc
	*/

	private void printStats(ISyncEvent evt)
	{
		Hashtable ht =  (Hashtable) evt.getEventInfo();
		if (ht == null)
			return;
		Enumeration enum = ht.keys();
		while (enum.hasMoreElements())
		{
			String key = (String) enum.nextElement();
			String val = (String) ht.get(key);
			logWrite(key + "=" + val);
		}
	}

	/*
		A general logging method to print time, % complete, type of event,
		subscription set or subscription information, and a message
	*/

	private void logISync(ISyncEvent evt) 
	{
		String buf = "";

		// Get the timestamp
		buf += new Timestamp(System.currentTimeMillis()) + " "
				+ evt.getSyncProgress() + "% ";
	
		// Get the error/status code			
		buf += "DSYC" + evt.getEventCode();
		if (evt.getEventType() == ISync.EVTTYPE_INFO)
			buf += "I:\t";
		else if (evt.getEventType() == ISync.EVTTYPE_RETRY)
			buf+= "R:\t";
		else if (evt.getEventType() == ISync.EVTTYPE_QUERY)
			buf+= "Q:\t";
		else if (evt.getEventType() == ISync.EVTTYPE_ERROR)
			buf+= "E:\t";
		else if (evt.getEventType() == ISync.EVTTYPE_CONFLICT)
			buf+= "C:\t";
		
		buf += eventCodetoMsg(evt.getEventCode());
	
		switch (evt.getEventCode())
			{
			case ISync.EVT_INF_SUBSSET_CANCELED :
			case ISync.EVT_INF_SUBSSET_FAILED :
			case ISync.EVT_INF_SUBSSET_STARTED :
			case ISync.EVT_INF_SUBSSET_SUCCEEDED :
				buf += ": " + evt.getSubscriptionSetName();
				break;

			case ISync.EVT_INF_SYNCING_SUBS :
				buf += ": " + evt.getSubscriptionName();
				break;

			default :
				break;
		}


		logWrite(buf);	
	}

	private void logWrite(String message)
	{
		System.out.println(message);
		logWriter.println(message);
	}

	private void logExceptionStack(Exception e)
	{
		e.printStackTrace();
		e.printStackTrace(logWriter);
	}

	/**
		Set stream writer
	*/

	public void setPrintStream(PrintStream os) 
	{
		logWriter = os;
	}

	/**
		Message for event code
	*/

	public static String eventCodetoMsg(int eventCode) 
	{
		switch(eventCode)
		{
		  case ISync.EVT_INF_GENERAL:
			  return "General information"; 
		  case ISync.EVT_INF_SYNC_STARTED:
			  return "Synchronization started ";
		  case ISync.EVT_INF_PREP_MSG:
			  return "Preparing message ";
		  case ISync.EVT_INF_SEND_MSG:
			  return "Sending message ";
		  case ISync.EVT_INF_WAIT_MSG:
			  return "Awaiting server reply ";
		  case ISync.EVT_INF_APPLY_MSG:
			  return "Applying server message"; 
		  case ISync.EVT_INF_CANCELING_SYNC:
			  return "Canceling synchronization ";
		  case ISync.EVT_INF_SYNC_PROG:
			  return "Synchronization progress ";
		  case ISync.EVT_INF_SUBSSET_STARTED:
			  return "Subscription set synchronization started ";
		  case ISync.EVT_INF_SYNCING_SUBS:
			  return "Synchronizing subscription ";
		  case ISync.EVT_INF_SUBSSET_FAILED:
			  return "Subscription set synchronization failed ";
		  case ISync.EVT_INF_SUBSSET_CANCELED:
			  return "Subscription set synchronization canceled by user ";
		  case ISync.EVT_INF_SUBSSET_SUCCEEDED:
			  return "Subscription set synchronization succeeded ";
		  case ISync.EVT_INF_SYNC_SUCCEEDED:
			  return "Synchronization succeeded ";
		  case ISync.EVT_INF_SYNC_FAILED:
			  return "Synchronization failed ";
		  case ISync.EVT_INF_SYNC_CANCELED:
			  return "Synchronization canceled ";
		  case ISync.EVT_INF_NO_NEW_CHANGE:
			  return "Info no new changes.";
		  case ISync.EVT_INF_LOGIN_FAILED:
			  return "Info login failed.";
		  case ISync.EVT_CFT_REJECT:
			  return "Info conflict reject.";
		  case ISync.EVT_ERR_OPEN_ADAPTER:
			  return "Error opening adapter.";
		  case ISync.EVT_ERR_LOAD_ADAPTER:
			  return "Error loading adapter.";
		  case ISync.EVT_ERR_CLOSE_ADAPTER:
			  return "Error closing adapter.";
		  case ISync.EVT_ERR_AUTHENTICATE_KEY:
			  return "Error authentication key.";
		  case ISync.EVT_ERR_CLIENT_CRYPTO_FAILED:
			  return "Error Client crypto failed.";
		  case ISync.EVT_ERR_ENCRYPT_NOT_AVAIL:
			  return "Error encryption not available.";
		  case ISync.EVT_ERR_ENCRYPT_LIB_OPEN:
			  return "Error opening encryption library.";
		  case ISync.EVT_ERR_SUBS_NOT_FOUND:
			  return "Error subscription not found.";
		  case ISync.EVT_ERR_SUBS_NOT_AVAIL:
			  return "Error subscription not available.";
		  case ISync.EVT_ERR_CLOSE_NET_LIB:
			  return "Error closing net lib.";
		  case ISync.EVT_ERR_OUT_OF_MEMORY:
			  return "Error out of memory!";
		  case ISync.EVT_ERR_ALLOC_RESOURCE:
			  return "Error allocating resource.";
		  case ISync.EVT_ERR_CONN_DATA_SRC:
			  return "Error connecting to data src.";
		  case ISync.EVT_ERR_DCONN_DATA_SRC:
			  return "Error DCONN_DATA_SRC?";
		  case ISync.EVT_ERR_NO_DATA:
			  return "Error no data.";
		  case ISync.EVT_ERR_INVALID_USER:
			  return "Error invalid user.";
		  case ISync.EVT_ERR_SYNC_DISABLED:
			  return "Error sync disabled.";
		  case ISync.EVT_ERR_SERVER_EXCEPTION:
			  return "Error server exception!!";
		  case ISync.EVT_ERR_MSG_FORMAT:
			  return "Error message format.";
		  case ISync.EVT_ERR_NOT_FOUND:
			  return "Error not found?";
		  case ISync.EVT_ERR_END_OF_DATA:
			  return "Error end of data.";
		  case ISync.EVT_ERR_DATA_TOO_LONG:
			  return "Error data too long!";
		  case ISync.EVT_ERR_READ_ONLY:
			  return "Error read only..";
		  case ISync.EVT_ERR_OPERATION:
			  return "Error operation.";
		  case ISync.EVT_ERR_UNAUTHORIZED:
			  return "Error unauthorized.";
		  case ISync.EVT_ERR_NOT_AVAIL:
			  return "Error not available.";
		  case ISync.EVT_ERR_INTERNAL:
			  return "Error Internal??";
		  case ISync.EVT_TRY_SERVER_BUSY:
			  return "Try to send request to busy server again.";
		  case ISync.EVT_TRY_NET_CONNECT:
			  return "Try network connection again.";
		  case ISync.EVT_TRY_NET_SEND:
			  return "Try to send request again.";
		  case ISync.EVT_TRY_NET_RECEIVE:
			  return "Try to receive reply again.";
		  case ISync.EVT_TRY_NET_TIMEOUT:
			  return "Timed out receiving, try again.";
		  case ISync.EVT_TRY_SERVER_AVAIL:
			  return "Try to contact the server again.";
		  case ISync.EVT_QUE_LOGIN:
			  return "Login information required by adapter.";
		  case ISync.EVT_QUE_CANCEL_UPON_ERROR:
			  return "Inquery if user wants to continue with the synchronization or cancel";
		  case ISync.EVT_QUE_SUBSCRIPTION_TARGET:
		  	  return "Query Per-subscription Target Dir";
		  case ISync.EVT_INF_STATS:
			  return "Performance statistics";
		  case ISync.EXC_NOT_SUPPORTED:
			  return "This feature is not supported.";
		  case ISync.EXC_NOT_EXIST:
			  return "Exception - the target/subscription set specified does not exist";
		  case ISync.EXC_PREF_UNKNOWN_ID:
			  return "Exception - unknown pref ID.";
		  case ISync.EXC_PREF_ERROR:
			  return "Exception - preference error.";
		  case ISync.EXC_TIMEOUT:
			  return "Exception - a time out occured.";
		  case ISync.EXC_ILLEGAL_ARG:
			  return "Exception - illegal argument.";
		  case ISync.EXC_ILLEGAL_STATE:
			  return "Exception - illegal state.";
		  case ISync.EXC_STORE_IN_USE:
			  return "Exception Store in use.";
		  case ISync.EXC_ILLEGAL_ACCESS:
			  return "Exception illegal access.";
		  case ISync.EXC_INVALID_FORMAT:
			  return "Exception invalid format.";
		  case ISync.EXC_CONFIG_ERROR:
			  return "Exception config error.";
		  case ISync.EXC_CONFIG_SYNC:
			  return "Exception config sync failed.";
		  case ISync.EXC_AUTH_FAILED:
			  return "Exception authentication failed.";
		  case ISync.EXC_INCOMP_VERSION:
			  return "Exception incompatible version.";
		  case ISync.EXC_INVALID_SESSION:
			  return "Exception invalid session.";
		  case ISync.EXC_SYNC_GROUP:
			  return "Exception user does not belong to any sync group.";
		  case ISync.EXC_REGISTER_DEVICE:
			  return "Exception failed to register device to user.";
		  case ISync.EXC_SERVER_FORBIDDEN:
			  return "Exception server forbidden.";
		  case ISync.EXC_SERVER_NOT_FOUND:
			  return "Exception server not found.";
		  case ISync.EXC_SERVER_ERROR:
			  return "Exception server error.";
		  case ISync.EXC_SERVER_NOT_AVAIL:
			  return "Exception server not available.";
		  case ISync.EXC_SERVER_BUSY:
			  return "Server busy.";
		  case ISync.EXC_NET_OPEN_CONN:
			  return "Exception network open connection.";
		  case ISync.EXC_NET_OPEN_LIB:
			  return "Exception opening net lib.";
		  case ISync.EXC_NET_CONNECT:
			  return "Error network connection.";
		  case ISync.EXC_NET_SEND:
			  return "Error network send request";
		  case ISync.EXC_NET_RECEIVE:
			  return "Error receiving reply.";
		  case ISync.EXC_NET_TIMEOUT:
			  return "Error receive timed out.";
		  case ISync.EXC_NET_RESOLVE_HOST:
			  return "Exception resolving host.";
		  case ISync.EXC_CREATE_IMAGE_WITHOUT_DEVICEID:
			  return "Exception attempt to create client image without a valid device ID.";
		  case ISync.EXC_NET_UNKNOWN:
			  return "Exception unknown network error.";

		  default:
			  return "Unknown event code " + eventCode;

		}
	}

	// Sample processes INSERT and UPDATE operations only
	private void processRejectedData(ConflictReader cr) 
	{
		int op;
		ResultSet pkrs = null;
		ResultSet rs = null;
		String cmsg;

		if (cr == null) 
			return;

		try
		{ 
			op = cr.getOperation();
			cmsg = "Conflicts detected in Table: " + cr.getTableName();
			logWrite(cmsg);

			rs = cr.getRejectedRows();
			processResultSet(rs);
		}
		catch (ISyncException iex) 
		{
			logWrite("ISyncException code: " + iex.getCode());
			logExceptionStack(iex);
		}
		catch (SQLException sqlex)
		{
			logExceptionStack(sqlex);

			while(sqlex != null)
			{
				logWrite("SQLERROR: \n" + sqlex.getErrorCode() + 
					", SQLState: " + sqlex.getSQLState() + 
					", Message: " + sqlex.getMessage() +
					", Vendor: " + sqlex.getErrorCode());
				sqlex = sqlex.getNextException();
			}
		}

		System.out.println("");
	} // processRejectedData

	private void processResultSet(ResultSet rs) throws SQLException 
	{
		ResultSetMetaData rsmd = rs.getMetaData();
		int numCols = rsmd.getColumnCount();
		String name;
		int type;
		int i = 0;

		while (rs.next()) 
		{
			i = 1;

			while (i <= numCols) 
			{
				name = rsmd.getColumnName(i);
				type = rsmd.getColumnType(i);
				System.out.print (name + ":");

				//BLOB - use getBytes()
				if (type == Types.LONGVARBINARY)
				{
					byte[] data = rs.getBytes(i);

					if (data == null || rs.wasNull()) 
					{
						System.out.print(" ( " + type + " ) ");
						System.out.println(" [ NULL ] ");
					}
					else 
					{
						System.out.print(" ( " + type + " ) ");
						System.out.println(" [ BLOB:" + data.length + " ] ");
					}
				}
				//Sample uses getString() for other types
				else 
				{
					String s = rs.getString(i);

					if (s == null || rs.wasNull()) 
					{
						System.out.print(" ( " + type + " ) ");
						System.out.println(" [ NULL ] ");
					}
					else 
					{
						System.out.print(" ( " + type + " ) ");
						System.out.println(" [ " + s + " ] ");
					}
				}

				i++;

			} //while numCols

		} //while rs.next

		rs.close();

	} //processResultSet
}
