/************************************************************************
 *
 * Licensed Materials - Property of IBM
 * Program Name: IBM DB2Everyplace Sync Client
 * (c ) Copyright IBM Corp. 1999-2003. All Rights Reserved.
 *
 * US Government Users Restricted Rights 
 * Use, duplication or disclosure restricted by GSA ADP Schedule
 * Contract with IBM Corp. Licensed Materials - Property of IBM
 *
 * IBM grants you ("Licensee") a non-exclusive, royalty free, license to
 * use Sample Code.  This software is provided as-is and without any
 * warranties, whether EXPRESS OR  IMPLIED, INCLUDING ANY IMPLIED
 * WARRANTY OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE OR 
 * NON-INFRINGEMENT. IBM AND ITS LICENSORS SHALL NOT BE LIABLE FOR ANY 
 * DAMAGES SUFFERED BY LICENSEE THAT RESULT FROM YOUR USE OF THE SOFTWARE.
 * IN NO EVENT WILL IBM OR ITS LICENSORS BE LIABLE FOR ANY LOST REVENUE,
 * PROFIT OR DATA, OR FOR DIRECT, INDIRECT, SPECIAL, CONSEQUENTIAL, 
 * INCIDENTAL OR PUNITIVE DAMAGES, HOWEVER CAUSED AND REGARDLESS OF THE 
 * THEORY OF LIABILITY, ARISING OUT OF THE USE OF OR INABILITY TO USE 
 * SOFTWARE, EVEN IF IBM HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH 
 * DAMAGES.
 *
 ************************************************************************
 */

using System;
using System.Data;				//IDataReader
using System.IO;				//Stream
using System.Collections;		//Hashtable

using IBM.Data.Sync;

#if (!UNDER_CF)
using IBM.Data.Sync.DB2e;		//standard framework
#else
using IBM.Data.Sync.DB2e.CF;	//compact framework
#endif

namespace DB2Sync
{
	/// <summary>
	/// Implementor of the EventIssued method which handles info, error,
	/// conflict and query types of events.
	/// </summary>
	public class DB2SyncListener
	{
		protected Hashtable isyncProps;
		protected bool retryMaxReached;
		protected StreamWriter logWriter;
 
		/// <summary>
		/// Create an instance setting in the isyncProps used
		/// to set up the sync engine and supply info in the EventIssued
		/// method.
		/// </summary>
		/// <param name="isyncProps">sync engine/event properties</param>
		public DB2SyncListener(Hashtable isyncProps) 
		{
			this.isyncProps = isyncProps;
		}

		/// <summary>
		/// Handle events from the isync driver which can be informational,
		/// error, data conflict, or query for information types of events.
		/// </summary>
		/// <param name="sender">sender</param>
		/// <param name="evt">event containing type, code, object for various use, etc.</param>
		/// <returns></returns>
		public int EventIssued(Object sender, ISyncEvent evt) 
		{
			if (evt == null) 
			{
				try
				{	
					if (logWriter != null)
						logWriter.WriteLine("Severe Error event is null on the callbacks from the iSyncEngine.");
				}
				catch(Exception e)
				{
					Console.WriteLine(e);
				}
				return ISync.RTNCB_DEFAULT;
			}

			int evtType = evt.GetEventType();
			int evtCode = evt.GetEventCode();

			LogISync(evt);

			switch (evtType) 
			{
				// display event status
				case ISync.EVTTYPE_INFO :
				case ISync.EVTTYPE_ERROR :
					if (evtCode == ISync.EVT_INF_STATS)
					{
						String trace = (String)isyncProps["isync.trace"];
						PrintStats(evt);
					}
					return ISync.RTNCB_DONE;

				case ISync.EVTTYPE_RETRY :
					if (evt.GetNumberOfRetries() >= 10)
						retryMaxReached = true;

					if(retryMaxReached)			
						return ISync.RTNCB_REPLY_NO;
					else
						return ISync.RTNCB_REPLY_YES;

				case ISync.EVTTYPE_CONFLICT :
					ConflictReader cr = (ConflictReader)evt.GetEventInfo();
					ProcessRejectedData(cr);
					return ISync.RTNCB_DONE;

				case ISync.EVTTYPE_QUERY :
					if (evtCode == ISync.EVT_QUE_LOGIN) 
					{
						String []login = (String[]) evt.GetEventInfo();
						try
						{
							if (logWriter != null)
								logWriter.WriteLine("Querying login for local encryption for database: " + login[0]);
						}
						catch(Exception e)
						{
							Console.WriteLine(e);
							LogWrite("Querying login for local encryption for database: " + login[0]);
						}

						if (isyncProps["target.db.user"] == null 
							|| isyncProps["target.db.password"] == null)
						{
							String buf;
							Console.Write("Username: ");
							buf = Console.ReadLine();
							login[1] = buf.Trim();
							Console.Write("Password: ");
							buf = Console.ReadLine();
							login[2] = buf.Trim();
						}
						else
						{
							login[1] = (String)isyncProps["target.db.user"];
							login[2] = (String)isyncProps["target.db.password"];
						}
						return ISync.RTNCB_DONE;
					}

					return ISync.RTNCB_DEFAULT;

					// ignore other event types
				default :
					break;
			}

			return ISync.RTNCB_DEFAULT;
		} // eventIssued

		/// <summary>
		/// Print sync statistics such as time, number of object synchronized,
		/// names of tables or files, etc
		/// </summary>
		/// <param name="evt"></param>
		private void PrintStats(ISyncEvent evt)
		{
				Hashtable ht =  (Hashtable) evt.GetEventInfo();
				if (ht == null)
					return;

				IDictionaryEnumerator e = ht.GetEnumerator();
				while (e.MoveNext())
					LogWrite((String) e.Key + "=" + (String) e.Value);
		} // PrintStats

		/// <summary>
		/// A general logging method to print time, % complete, type of event,
		/// subscription set or subscription information, and a message
		/// </summary>
		/// <param name="evt"></param>
		private void LogISync(ISyncEvent evt) 
		{
			String buf = "";

			// Get the timestamp
			buf += DateTime.Now.ToShortDateString() + " " 
				+ DateTime.Now.ToLongTimeString() + " "
				+ evt.GetSyncProgress() + "% ";
	
			// Get the error/status code			
			buf += "DSYC" + evt.GetEventCode();
			if(evt.GetEventType() == ISync.EVTTYPE_INFO)
				buf += "I:\t";
			else if(evt.GetEventType() == ISync.EVTTYPE_RETRY)
				buf+= "R:\t";
			else if(evt.GetEventType() == ISync.EVTTYPE_QUERY)
				buf+= "Q:\t";
			else if(evt.GetEventType() == ISync.EVTTYPE_ERROR)
				buf+= "E:\t";
			else if(evt.GetEventType() == ISync.EVTTYPE_CONFLICT)
				buf+= "C:\t";
		
			buf += EventCodetoMsg(evt.GetEventCode());
	
			switch (evt.GetEventCode())
			{
				case ISync.EVT_INF_SUBSSET_CANCELED :
				case ISync.EVT_INF_SUBSSET_FAILED :
				case ISync.EVT_INF_SUBSSET_STARTED :
				case ISync.EVT_INF_SUBSSET_SUCCEEDED :
					buf += (": " + evt.GetSubscriptionSetName());
					break;

				case ISync.EVT_INF_SYNCING_SUBS :
					buf += (": " + evt.GetSubscriptionName());
					break;

				default :
					break;
			}

			try
			{
				LogWrite(buf);	
			}
			catch(Exception ioe)
			{
				Console.WriteLine(ioe);
			}
		} // LogISync

		private void LogWrite(String message) //throws IOException
		{
			try
			{
				Console.WriteLine(message);
				if (logWriter != null)
					logWriter.WriteLine(message);
			}
			catch(Exception e)
			{
				Console.WriteLine(e);
			}
		} // LogWrite

		/// <summary>
		/// Set stream writer
		/// </summary>
		/// <param name="os"></param>
		public void SetOutPutStream(StreamWriter os) 
		{
			logWriter = os;
		}

		private static String EventCodetoMsg(int eventCode) 
		{
			switch(eventCode)
			{
				case ISync.EVT_INF_GENERAL: return "General information"; 
				case ISync.EVT_INF_SYNC_STARTED: return "Synchronization started ";
				case ISync.EVT_INF_PREP_MSG: return "Preparing message ";
				case ISync.EVT_INF_SEND_MSG: return "Sending message ";
				case ISync.EVT_INF_WAIT_MSG: return "Awaiting server reply ";
				case ISync.EVT_INF_APPLY_MSG: return "Applying server message"; 
				case ISync.EVT_INF_CANCELING_SYNC: return "Canceling synchronization ";
				case ISync.EVT_INF_SYNC_PROG: return "Synchronization progress ";
				case ISync.EVT_INF_SUBSSET_STARTED: return "Subscription Set synchronization started ";
				case ISync.EVT_INF_SYNCING_SUBS: return "Synchronizing subscription ";
				case ISync.EVT_INF_SUBSSET_FAILED: return "Subscription Set synchronization failed ";
				case ISync.EVT_INF_SUBSSET_CANCELED: return "Subscription Set synchronization canceled by user ";
				case ISync.EVT_INF_SUBSSET_SUCCEEDED: return "Subscription Set synchronization succeeded ";
				case ISync.EVT_INF_SYNC_SUCCEEDED: return "Synchronization succeeded ";
				case ISync.EVT_INF_SYNC_FAILED: return "Synchronization failed ";
				case ISync.EVT_INF_SYNC_CANCELED: return "Synchronization canceled ";
				case ISync.EVT_INF_NO_NEW_CHANGE: return "Info no new changes.";
				case ISync.EVT_INF_LOGIN_FAILED: return "Info login failed.";
				case ISync.EVT_CFT_REJECT: return "Info conflict reject.";
				case ISync.EVT_ERR_OPEN_ADAPTER: return "Error opening adapter.";
				case ISync.EVT_ERR_LOAD_ADAPTER: return "Error loading adapter.";
				case ISync.EVT_ERR_CLOSE_ADAPTER: return "Error closing adapter.";
				case ISync.EVT_ERR_AUTHENTICATE_KEY: return "Error authentication key.";
				case ISync.EVT_ERR_CLIENT_CRYPTO_FAILED: return "Error Client crypto failed.";
				case ISync.EVT_ERR_ENCRYPT_NOT_AVAIL: return "Error encryption not available.";
				case ISync.EVT_ERR_ENCRYPT_LIB_OPEN: return "Error opening encryption library.";
				case ISync.EVT_ERR_SUBS_NOT_FOUND: return "Error subscription not found.";
				case ISync.EVT_ERR_SUBS_NOT_AVAIL: return "Error subscription not available.";
				case ISync.EVT_ERR_CLOSE_NET_LIB: return "Error closing net lib.";
				case ISync.EVT_ERR_OUT_OF_MEMORY: return "Error out of memory!";
				case ISync.EVT_ERR_ALLOC_RESOURCE: return "Error allocating resource.";
				case ISync.EVT_ERR_CONN_DATA_SRC: return "Error connecting to data src.";
				case ISync.EVT_ERR_DCONN_DATA_SRC: return "Error DCONN_DATA_SRC?";
				case ISync.EVT_ERR_NO_DATA: return "Error no data.";
				case ISync.EVT_ERR_INVALID_USER: return "Error invalid user.";
				case ISync.EVT_ERR_SYNC_DISABLED: return "Error sync disabled.";
				case ISync.EVT_ERR_SERVER_EXCEPTION: return "Error server exception!!";
				case ISync.EVT_ERR_MSG_FORMAT: return "Error message format.";
				case ISync.EVT_ERR_NOT_FOUND: return "Error not found?";
				case ISync.EVT_ERR_END_OF_DATA: return "Error end of data.";
				case ISync.EVT_ERR_DATA_TOO_LONG: return "Error data too long!";
				case ISync.EVT_ERR_READ_ONLY: return "Error read only..";
				case ISync.EVT_ERR_OPERATION: return "Error operation.";
				case ISync.EVT_ERR_UNAUTHORIZED: return "Error unauthorized.";
				case ISync.EVT_ERR_NOT_AVAIL: return "Error not available.";
				case ISync.EVT_ERR_INTERNAL: return "Error Internal??";
				case ISync.EVT_TRY_SERVER_BUSY: return "Try to send request to busy server again.";
				case ISync.EVT_TRY_NET_CONNECT: return "Try network connection again.";
				case ISync.EVT_TRY_NET_SEND: return "Try to send request again.";
				case ISync.EVT_TRY_NET_RECEIVE: return "Try to receive reply again.";
				case ISync.EVT_TRY_NET_TIMEOUT: return "Timed out receiving, try again.";
				case ISync.EVT_TRY_SERVER_AVAIL: return "Try to contact the server again.";
				case ISync.EVT_QUE_LOGIN: return "Login information required by adapter.";
				case ISync.EVT_QUE_CANCEL_UPON_ERROR: return "Inquery if user wants to continue with the synchronization or cancel";
				case ISync.EVT_QUE_SUBSCRIPTION_TARGET:
					return "Query Per-subscription Target Dir";
				case ISync.EVT_INF_STATS:
					return "Performance statistics";
				case ISync.EXC_NOT_SUPPORTED: return "This feature is not supported.";
				case ISync.EXC_NOT_EXIST: return "Exception - the target/subscription Set specified does not exist";
				case ISync.EXC_PREF_UNKNOWN_ID: return "Exception - unknown pref ID.";
				case ISync.EXC_PREF_ERROR: return "Exception - preference error.";
				case ISync.EXC_TIMEOUT: return "Exception - a time out occured.";
				case ISync.EXC_ILLEGAL_ARG: return "Exception - illegal argument.";
				case ISync.EXC_ILLEGAL_STATE: return "Exception - illegal state.";
				case ISync.EXC_STORE_IN_USE: return "Exception Store in use.";
				case ISync.EXC_ILLEGAL_ACCESS: return "Exception illegal access.";
				case ISync.EXC_INVALID_FORMAT: return "Exception invalid format.";
				case ISync.EXC_CONFIG_ERROR: return "Exception config error.";
				case ISync.EXC_CONFIG_SYNC: return "Exception config sync failed.";
				case ISync.EXC_AUTH_FAILED: return "Exception authentication failed.";
				case ISync.EXC_INCOMP_VERSION: return "Exception incompatible version.";
				case ISync.EXC_INVALID_SESSION: return "Exception invalid session.";
				case ISync.EXC_SYNC_GROUP: return "Exception user does not belong to any sync group.";
				case ISync.EXC_REGISTER_DEVICE: return "Exception failed to register device to user.";
				case ISync.EXC_SERVER_FORBIDDEN: return "Exception server forbidden.";
				case ISync.EXC_SERVER_NOT_FOUND: return "Exception server not found.";
				case ISync.EXC_SERVER_ERROR: return "Exception server error.";
				case ISync.EXC_SERVER_NOT_AVAIL: return "Exception server not available.";
				case ISync.EXC_SERVER_BUSY: return "Exception server busy.";
				case ISync.EXC_NET_OPEN_CONN: return "Exception network open connection.";
				case ISync.EXC_NET_OPEN_LIB: return "Exception opening net lib.";
				case ISync.EXC_NET_CONNECT: return "Exception network connection.";
				case ISync.EXC_NET_SEND: return "Exception network send request.";
				case ISync.EXC_NET_RECEIVE: return "Exception network receive reply.";
				case ISync.EXC_NET_TIMEOUT: return "Exception network timeout.";
				case ISync.EXC_NET_RESOLVE_HOST: return "Exception resolving host.";
		  		case ISync.EXC_CREATE_IMAGE_WITHOUT_DEVICEID: return "Exception attempt to create client image without a valid device ID.";
				case ISync.EXC_NET_UNKNOWN: return "Exception unknown network error.";

				default:
					return "Unknown event code " + eventCode;
			}
	
		} // EventCodetoMsg

		// process INSERT and UPDATE operation only	
		private void ProcessRejectedData(ConflictReader cr) 
		{
			int op;
			IDataReader rs = null;
			String cmsg;

			if (cr == null)
				return;

			try 
			{ 
				op = cr.GetOperation();
				cmsg = "Conflicts detected in Table: " + cr.GetTableName();
         
				Console.WriteLine (cmsg);

				rs = cr.GetRejectedRows();
				ProcessResultSet(rs);
			}
			catch (ISyncException iex) 
			{
				Console.WriteLine (iex.GetCode());
				Console.WriteLine (iex.ToString());
			}
			catch (DB2eSQLException sqlex) 
			{
				Console.WriteLine(sqlex.ToString());
				Console.WriteLine("SQLERROR: \n" + sqlex.GetErrorCode() + 
					", SQLState: " + sqlex.GetSQLState() + 
					", Message: " + sqlex.Message +
					", Vendor: " + sqlex.GetErrorCode() );
			}

			Console.WriteLine("");
		} //ProcessRejectedData

		private void ProcessResultSet(IDataReader rs)
		{
			DataTable rsmd = rs.GetSchemaTable();
			int numCols = rs.FieldCount;
			String name;
			Type type;
			int i = 0;

			while (rs.Read()) 
			{
				while (i < numCols) 
				{
					name = rs.GetName(i);
					type = rs.GetFieldType(i);
					Console.Write (name + ":");

					if (type == typeof(Byte []))
					{
						if (rs.IsDBNull(i)) 
						{
							Console.Write(" ( " + type + " ) ");
							Console.Write(" [ NULL ] ");
						}
						else 
						{
							int numBytes = (int) rs.GetBytes(i, 0, null, 0, 0);
							byte [] data = new byte[numBytes];
							rs.GetBytes(i, 0, data, 0, numBytes);

							Console.Write(" ( " + type + " ) ");
							Console.Write(" [ BLOB:" + data.Length + " ] ");
						}
					}
					else if (rs.GetDataTypeName(i).Equals("Date") ||
						rs.GetDataTypeName(i).Equals("DateTime"))
					{
						if (rs.IsDBNull(i)) 
						{
							Console.Write(" ( " + type + " ) ");
							Console.Write(" [ NULL ] ");
						}
						else 
						{
							DateTime s = rs.GetDateTime(i);

							Console.Write(" ( " + type + " ) ");
							Console.Write(" [ " + s + " ] ");
						}
					}
					else 
					{
						if (rs.IsDBNull(i)) 
						{
							Console.Write(" ( " + type + " ) ");
							Console.Write(" [ NULL ] ");
						}
						else 
						{
							String s = rs.GetString(i);

							Console.Write(" ( " + type + " ) ");
							Console.Write(" [ " + s + " ] ");
						}
					}

					i++;
				}                
			}

			rs.Close();
		} // ProcessResultSet

	} //class DB2SyncListener

} //namespace DB2Sync
